/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/Envelope.cpp

  \brief An Envelope defines a 2D rectangular region.
*/

// TerraLib
#include "../BuildConfig.h"
#include "../common/Exception.h"
#include "../core/translator/Translator.h"
#include "../srs/Converter.h"
#include "Coord2D.h"
#include "Envelope.h"
#include "Exception.h"
#include "LineString.h"
#include "Polygon.h"

// STL
#include <algorithm>
#include <cassert>
#include <cmath>
#include <memory>

te::gm::Polygon CreatePolygon(const te::gm::Envelope& box, const int& srid, const int& numberOfIntermediateCoords);

te::gm::LineString CreateLine(const te::gm::Coord2D& startCoord, const te::gm::Coord2D& endCoord, const int& srid, const int& numberOfIntermediateCoords);


te::gm::Coord2D te::gm::Envelope::getLowerLeft() const
{
  return Coord2D(m_llx, m_lly);
}

te::gm::Coord2D te::gm::Envelope::getUpperRight() const
{
  return Coord2D(m_urx, m_ury);
}

te::gm::Coord2D te::gm::Envelope::getCenter() const
{
  return Coord2D(m_llx + (m_urx - m_llx) / 2.0, m_lly + (m_ury - m_lly) / 2.0);
}

double te::gm::Envelope::distance(const Envelope& rhs) const
{
  if(intersects(rhs))
    return 0.0;

  double dx = 0.0;

// is this envelope to the left?
  if(m_urx < rhs.m_llx)
    dx = rhs.m_llx - m_urx;

// is this envelope to the right?
  if(m_llx > rhs.m_urx)
    dx = m_llx - rhs.m_urx;

  double dy = 0.0;

// is this envelope below?
  if(m_ury < rhs.m_lly)
    dy = rhs.m_lly - m_ury;

// is this envelope above?
  if(m_lly > rhs.m_ury)
    dy = m_lly - rhs.m_ury;

// are the envelopes aligned horizontally?
  if(dx == 0.0)
    return dy;
  
// are the envelopes aligned vertically?
  if(dy == 0.0)
    return dx;

  return sqrt((dx * dx) + (dy * dy));
}

void te::gm::Envelope::transform(int oldsrid, int newsrid)
{
#ifdef TERRALIB_MOD_SRS_ENABLED

  if (oldsrid == newsrid)
    return;

  te::gm::Envelope env(*this);

  te::gm::Polygon polygon = CreatePolygon(env, oldsrid, 32);

  polygon.transform(newsrid);

  const te::gm::Envelope* transEnv = polygon.getMBR();

  m_llx = transEnv->m_llx;
  m_urx = transEnv->m_urx;
  m_lly = transEnv->m_lly;
  m_ury = transEnv->m_ury;

#else
  throw Exception(TE_TR("transform method is not supported!"));
#endif // TERRALIB_MOD_SRS_ENABLED
}

te::gm::Polygon CreatePolygon(const te::gm::Envelope& box, const int& srid, const int& numberOfIntermediateCoords)
{
  te::gm::Coord2D lowerLeft(box.m_llx, box.m_lly);
  te::gm::Coord2D lowerRight(box.m_urx, box.m_lly);

  te::gm::Coord2D upperLeft(box.m_llx, box.m_ury);
  te::gm::Coord2D upperRight(box.m_urx, box.m_ury);

  te::gm::LineString bottomLine = CreateLine(lowerLeft, lowerRight, srid, numberOfIntermediateCoords);
  te::gm::LineString leftLine = CreateLine(lowerRight, upperRight, srid, numberOfIntermediateCoords);
  te::gm::LineString topLine = CreateLine(upperRight, upperLeft, srid, numberOfIntermediateCoords);
  te::gm::LineString rightLine = CreateLine(upperLeft, lowerLeft, srid, numberOfIntermediateCoords);

  std::vector<te::gm::Coord2D> coords;

  //we do not add the last coord to avoid duplicity
  for (std::size_t i = 0; i < bottomLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(bottomLine.getX(i), bottomLine.getY(i)));

  for (std::size_t i = 0; i < leftLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(leftLine.getX(i), leftLine.getY(i)));

  for (std::size_t i = 0; i < topLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(topLine.getX(i), topLine.getY(i)));

  for (std::size_t i = 0; i < rightLine.size() - 1; ++i)
    coords.push_back(te::gm::Coord2D(rightLine.getX(i), rightLine.getY(i)));

  te::gm::LineString* fullLine = new te::gm::LineString(coords.size(), te::gm::LineStringType);

  for (std::size_t t = 0; t < coords.size(); ++t)
    fullLine->setPoint(t, coords[t].getX(), coords[t].getY());

  te::gm::Polygon polygon(1, te::gm::PolygonType, srid);
  polygon.setRingN(0, fullLine);

  return polygon;
}

te::gm::LineString CreateLine(const te::gm::Coord2D& startCoord, const te::gm::Coord2D& endCoord, const int& srid, const int& numberOfIntermediateCoords)
{
  double gapX = (endCoord.getX() - startCoord.getX()) / (double)(numberOfIntermediateCoords + 1);
  double gapY = (endCoord.getY() - startCoord.getY()) / (double)(numberOfIntermediateCoords + 1);

  std::vector<te::gm::Coord2D> coords;

  coords.push_back(startCoord);
  for (std::size_t i = 0; i < numberOfIntermediateCoords; ++i)
  {
    std::size_t coordIndex = i + 1;
    te::gm::Coord2D coord(startCoord.getX() + (coordIndex * gapX), startCoord.getY() + (coordIndex * gapY));
    coords.push_back(coord);
  }
  coords.push_back(endCoord);

  te::gm::LineString resultLine(coords.size(), te::gm::LineStringType, srid);

  for (std::size_t t = 0; t < coords.size(); ++t)
    resultLine.setPoint(t, coords[t].getX(), coords[t].getY());

  return resultLine;
}

